/*
 * This file is based on CommandLineToArgvW from the Wine project, 
 * version 1.1.24.
 *
 * Copyright 1998 Marcus Meissner
 * Copyright 1998 Juergen Schmied (jsch)  *  <juergen.schmied@metronet.de>
 * Copyright 2010 David Goodwin <david@zx.net.nz>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <windows.h>
#include "CmdLineToArgvA.h"

/*************************************************************************
 * CommandLineToArgvA
 *
 * We must interpret the quotes in the command line to rebuild the argv
 * array correctly:
 * - arguments are separated by spaces or tabs
 * - quotes serve as optional argument delimiters
 *   '"a b"'   -> 'a b'
 * - escaped quotes must be converted back to '"'
 *   '\"'      -> '"'
 * - an odd number of '\'s followed by '"' correspond to half that number
 *   of '\' followed by a '"' (extension of the above)
 *   '\\\"'    -> '\"'
 *   '\\\\\"'  -> '\\"'
 * - an even number of '\'s followed by a '"' correspond to half that number
 *   of '\', plus a regular quote serving as an argument delimiter (which
 *   means it does not appear in the result)
 *   'a\\"b c"'   -> 'a\b c'
 *   'a\\\\"b c"' -> 'a\\b c'
 * - '\' that are not followed by a '"' are copied literally
 *   'a\b'     -> 'a\b'
 *   'a\\b'    -> 'a\\b'
 *
 * Note:
 * '\t' == 0x09
 * ' '  == 0x20
 * '"'  == 0x22
 * '\\' == 0x5c
 */
 LPSTR* CommandLineToArgvA(LPCSTR lpCmdline, int* numargs)
 {
     DWORD argc;
     LPSTR  *argv;
     LPCSTR cs;
     LPSTR arg,s,d;
     LPSTR cmdline;
     int in_quotes,bcount;
 
     if (*lpCmdline==0)
     {
         /* Return the path to the executable */
         DWORD len, size=16;
 
         argv=(LPSTR*)LocalAlloc(LMEM_FIXED, size);
         for (;;)
         {
             len = GetModuleFileNameA(0, (LPSTR)(argv+1), (size-sizeof(LPSTR))/sizeof(CHAR));
             if (!len)
             {
                 LocalFree(argv);
                 return NULL;
             }
             if (len < size) break;
             size*=2;
             argv=(LPSTR*)LocalReAlloc(argv, size, 0);
         }
         argv[0]=(LPSTR)(argv+1);
         if (numargs)
             *numargs=1;
 
         return argv;
     }
 
     /* to get a writable copy */
     argc=0;
     bcount=0;
     in_quotes=0;
     cs=lpCmdline;
     while (1)
     {
         if (*cs==0 || ((*cs==0x09 || *cs==0x20) && !in_quotes))
         {
             /* space */
             argc++;
             /* skip the remaining spaces */
             while (*cs==0x09 || *cs==0x20) {
                 cs++;
             }
             if (*cs==0)
                 break;
             bcount=0;
             continue;
         }
         else if (*cs==0x5c)
         {
             /* '\', count them */
             bcount++;
         }
         else if ((*cs==0x22) && ((bcount & 1)==0))
         {
             /* unescaped '"' */
             in_quotes=!in_quotes;
             bcount=0;
         }
         else
         {
             /* a regular character */
             bcount=0;
         }
         cs++;
     }
     /* Allocate in a single lump, the string array, and the strings that go with it.
      * This way the caller can make a single GlobalFree call to free both, as per MSDN.
      */
     argv=(LPSTR*)LocalAlloc(LMEM_FIXED, argc*sizeof(LPSTR)+(strlen(lpCmdline)+1)*sizeof(CHAR));
     if (!argv)
         return NULL;
     cmdline=(LPSTR)(argv+argc);
     strcpy(cmdline, lpCmdline);

     argc=0;
     bcount=0;
     in_quotes=0;
     arg=d=s=cmdline;
     while (*s)
     {
         if ((*s==0x09 || *s==0x20) && !in_quotes)
         {
             /* Close the argument and copy it */
             *d=0;
             argv[argc++]=arg;

             /* skip the remaining spaces */
             do {
                 s++;
             } while (*s==0x09 || *s==0x20);

             /* Start with a new argument */
             arg=d=s;
             bcount=0;
         }
         else if (*s==0x5c)
         {
             /* '\\' */
             *d++=*s++;
             bcount++;
         }
         else if (*s==0x22)
         {
             /* '"' */
             if ((bcount & 1)==0)
             {
                 /* Preceded by an even number of '\', this is half that
                  * number of '\', plus a quote which we erase.
                  */
                 d-=bcount/2;
                 in_quotes=!in_quotes;
                 s++;
             }
             else
             {
                 /* Preceded by an odd number of '\', this is half that
                  * number of '\' followed by a '"'
                  */
                 d=d-bcount/2-1;
                 *d++='"';
                 s++;
             }
             bcount=0;
         }
         else
         {
             /* a regular character */
             *d++=*s++;
             bcount=0;
         }
     }
     if (*arg)
     {
         *d='\0';
         argv[argc++]=arg;
     }
     if (numargs)
         *numargs=argc;
 
     return argv;
 }
