/*****************************************************************************
 * Application Name: SIMG32 - DIMG 32bit Simple Imager for Windows
 *        Copyright: (C) Copyright David Goodwin, 2010
 *          License: GNU GPL v2
 *****************************************************************************
	This file is part of "WinDIMG32".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * Ensure _UNICODE is defined if UNICODE is. UNICODE affets the windows header
 * functions (maps MessageBox to MessageBoxW) where as _UNICODE affects the
 * C runtime headers and stuff like tchar.h.
 */
#ifdef UNICODE
#ifndef _UNICODE
#define _UNICODE
#endif
#endif

/*
 * Standard Headers
 */
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <tchar.h>


/*
 * Header for this file
 */
#include "imager.h"
#include "resource.h"

/*
 * Library Headers
 */
#include "../vam/winvam32_dll.h"

/*
 * Constants
 */
#define MAX_MESSAGE_LEN 80
#define PROMPT_MESSAGE_LEN 255

/*
 * local variables
 */
static bool imagingCanceled;
static char targetDrive;

/*
 * Forward declaration
 */
LRESULT CALLBACK ProgressDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK InsertPromptDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK ConfirmDlgProc(HWND, UINT, WPARAM, LPARAM);
bool doReadLoop(HWND hwnd, TCHAR fileName[]);
bool doWriteLoop(HWND hwnd, TCHAR fileName[]);
void closeFile(FILE* file);
void showErrorMessage(int code);

/*
 * Does the imaging. Terrible name but descriptive enough.
 */
int doImaging(char driveLetter,      /* Drive letter (A,B,etc) */
			  int imageMode,	     /* Image Mode (1,2) */
			  bool noConfirm,		 /* Suppress confirm write? */
			  bool noCancel,		 /* disable cancel button? */
			  TCHAR fileName[],      /* Image file name */
			  HINSTANCE hInstance) { /* Program Instance */
	
	HWND hwnd;
	int err;
	bool success;
    char message[MAX_MESSAGE_LEN];
	long dlgResult;

	targetDrive = driveLetter;

	memset(message,0,MAX_MESSAGE_LEN);

	hwnd = CreateDialog(hInstance,	
						MAKEINTRESOURCE(IDD_PROGRESS),
						NULL,
						reinterpret_cast<DLGPROC>(ProgressDlgProc));

	if (noCancel)
		EnableWindow(GetDlgItem(hwnd,IDCANCEL),FALSE);

	ShowWindow(hwnd,SW_SHOWDEFAULT);
	UpdateWindow(hwnd);

    sprintf(message,"Setting up...");
	SendDlgItemMessage(hwnd,
		IDC_STATUS,
		WM_SETTEXT,
		0,
		(LPARAM)(LPCTSTR)&message);

	sprintf(message,"Drive %c:",driveLetter);
	SendDlgItemMessage(hwnd,
		IDC_DRIVE,
		WM_SETTEXT,
		0,
		(LPARAM)(LPCTSTR)&message);

	if (imageMode == IMGMODE_WRITE && !noConfirm) {

		dlgResult = DialogBox(hInstance,
							  MAKEINTRESOURCE(IDD_INSERT_DISK),
							  NULL,
							  reinterpret_cast<DLGPROC>(ConfirmDlgProc));
		if (dlgResult == FALSE) {
			// The user has given up
			return IMG_CANCELED;
		}
	}

	err = VAMERR_FAILURE;
	while (err != VAMERR_SUCCESS) {
		// Prepare to do some imaging (open the drive, etc).
		err = vam_open_volume(driveLetter);
		
		if (err != VAMERR_SUCCESS) {

			if (err == VAMERR_OPEN_FAILED) {
				dlgResult = DialogBox(hInstance,
									  MAKEINTRESOURCE(IDD_INSERT_DISK),
									  NULL,
									  reinterpret_cast<DLGPROC>(InsertPromptDlgProc));
				if (dlgResult == FALSE) {
					// The user has given up.
					return IMG_CANCELED;
				}
				
			} else {
				showErrorMessage(err);
				return IMG_FAILED;
			}
		}
	}

	if (imageMode == IMGMODE_READ) {
		success = doReadLoop(hwnd,fileName);
	} else if (imageMode == IMGMODE_WRITE) {
		success = doWriteLoop(hwnd,fileName);
	}

	err = vam_close_volume();

	if (err != VAMERR_SUCCESS) {
		showErrorMessage(err);
		return IMG_FAILED;
	}

	if (imagingCanceled)
		return IMG_CANCELED;
	else {
		if (success)
			return IMG_SUCCESS;
		else
			return IMG_FAILED;
	}
}

bool doReadLoop(HWND hwnd, TCHAR *fileName) {
	MSG msg;
	vam_diskgeom geom;
	long totalTracks, currentTrack, trackSize;
	bool imagingComplete;
	FILE *file;
	int errcode;
	char* buffer = NULL;
	int lengthRead, lengthWritten;
	LPARAM progressRange;
    char message[MAX_MESSAGE_LEN];
	memset(message,0,MAX_MESSAGE_LEN);

	// Get the disk geometry
	errcode = vam_read_diskgeom();
	geom = vam_get_geometry();
	totalTracks = geom.cylinders * geom.heads;
	trackSize = geom.sector_size * geom.sectors;
	currentTrack = 0;

	// Setup the UI a bit
	progressRange = MAKELPARAM(0,totalTracks);
	SendDlgItemMessage(hwnd,
		IDC_PROGRESS,
		PBM_SETRANGE,
		0,
		progressRange);

	// Create/Open the target file
    file = _tfopen(fileName,TEXT("wb"));

	if(file == NULL) {
		// Failed to open file
		MessageBox(NULL,TEXT("Failed to open target image file."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// Allocate a buffer
	buffer = (char*)malloc(trackSize);

	if (buffer == NULL) {
		MessageBox(NULL,TEXT("Insufficient memory to allocate read buffer."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// The event / imaging loop
	imagingComplete = false;
	imagingCanceled = false;
	while (!imagingComplete && !imagingCanceled) {
		// Do some imaging

		// Read in the current track
		errcode = vam_read_data(currentTrack,
								buffer,
								trackSize,
								&lengthRead);

		if (errcode != VAMERR_SUCCESS || lengthRead < trackSize) {
			// Read failed. Just give up for now.
			showErrorMessage(errcode);
			free(buffer);
			closeFile(file);
			return false;
		}


		// Write it out to the image file
		lengthWritten = fwrite(buffer,sizeof(char),trackSize, file);

		// Move on to the next track
		currentTrack = currentTrack + 1;
		if (currentTrack == totalTracks)
                        imagingComplete = true;

		// Update the UI
        sprintf(message,"Reading Track %d/%d",(int)currentTrack,(int)totalTracks);
		SendDlgItemMessage(hwnd,
			IDC_STATUS,
			WM_SETTEXT,
			0,
			(LPARAM)(LPCTSTR)&message);

		SendDlgItemMessage(hwnd,
			IDC_PROGRESS,
			PBM_SETPOS,
			currentTrack,
			0);

		// Process some events
		while (PeekMessage(&msg,hwnd,0,0,PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Deallocate the buffer & close the file
	free(buffer);
	closeFile(file);

	return true;
}

bool doWriteLoop(HWND hwnd, TCHAR fileName[]) {
	MSG msg;
	vam_diskgeom geom;
	unsigned long totalTracks, currentTrack, trackSize;
	bool imagingComplete;
	FILE *file;
	int errcode;
	char* buffer = NULL;
	unsigned int lengthRead;
	LPARAM progressRange;
    char message[MAX_MESSAGE_LEN];
	DWORD lengthWritten;

	memset(message,0,MAX_MESSAGE_LEN);

	// Open source file
	file = _tfopen(fileName,TEXT("rb"));
	if (file == NULL) {
		// Failed to open file
		MessageBox(NULL,TEXT("Failed to open source image file."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// Allocate a temporary buffer
	trackSize = 1024;
	buffer = (char*)malloc(trackSize);

	if (buffer == NULL) {
		MessageBox(NULL,TEXT("Insufficient memory to allocate read buffer."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// Read in the boot sector
	lengthRead = fread(buffer,sizeof(char),trackSize,file);

	if (lengthRead < trackSize) {
		MessageBox(NULL,TEXT("Failed to read first sector from image file."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// Initialise the geometry structure and grab a copy.
	vam_set_geometry_from_data(buffer,trackSize);
	geom = vam_get_geometry();

	// Do some calculations and create the buffer
	free(buffer);
	buffer = NULL;
	totalTracks = geom.cylinders * geom.heads;
	trackSize = geom.sector_size * geom.sectors;
	buffer = (char*)malloc(trackSize);

	if (buffer == NULL) {
		MessageBox(NULL,TEXT("Insufficient memory to allocate read buffer."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	// Back to the start of the image file
	if (fseek(file,0,SEEK_SET) != 0) {
		MessageBox(NULL,TEXT("Failed to seek to beginning of image file."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return false;
	}

	currentTrack = 0;

	// Setup the UI a bit
	progressRange = MAKELPARAM(0,totalTracks);
	SendDlgItemMessage(hwnd,
		IDC_PROGRESS,
		PBM_SETRANGE,
		0,
		progressRange);

	// The event / imaging loop
	imagingComplete = false;
	imagingCanceled = false;
	while (!imagingComplete && !imagingCanceled) {

		// Read in a track from the file
		lengthRead = fread(buffer,sizeof(char),trackSize,file);

		if (lengthRead < trackSize) {
			MessageBox(NULL,TEXT("Failed to read from image file."),
					TEXT("Error"), MB_OK | MB_ICONSTOP);
			return false;
		}


		// write the track out to disk
		errcode = vam_write_data(currentTrack,
								 buffer,
								 trackSize,
								 &lengthWritten);

		if (errcode != VAMERR_SUCCESS || lengthWritten < trackSize) {
			// write failed. Just give up for now.
			showErrorMessage(errcode);
			free(buffer);
			closeFile(file);
			return false;
		}

		// Move on to the next track
		currentTrack = currentTrack + 1;
		if (currentTrack == totalTracks)
                        imagingComplete = true;

		// Update the UI
        sprintf(message,"Writing Track %d/%d",(int)currentTrack,(int)totalTracks);
		SendDlgItemMessage(hwnd,
			IDC_STATUS,
			WM_SETTEXT,
			0,
			(LPARAM)(LPCTSTR)&message);

		SendDlgItemMessage(hwnd,
			IDC_PROGRESS,
			PBM_SETPOS,
			currentTrack,
			0);

		// Process some events
		while (PeekMessage(&msg,hwnd,0,0,PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Deallocate the buffer & close the file
	free(buffer);
	closeFile(file);

	return true;
}

void closeFile(FILE* file) {
	if (fclose(file)) {
		// Failed to close file.
		MessageBox(NULL,TEXT("Failed to close image file."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
	}
}

/*
 * The dialog event processing function
 */
LRESULT CALLBACK ProgressDlgProc(HWND hwnd,
								 UINT iMsg,
								 WPARAM wParam,
								 LPARAM lParam) {

	switch(iMsg)
	{
	case WM_INITDIALOG:

		return TRUE;

	case WM_COMMAND:
		switch(wParam)
		{
			case IDCANCEL:
				imagingCanceled = true;
				DestroyWindow(hwnd);
				return TRUE;
		}
		break;
	}

	return FALSE;
}

/*
 * The dialog event processing function
 */
LRESULT CALLBACK InsertPromptDlgProc(HWND hwnd,
							 UINT iMsg,
							 WPARAM wParam,
							 LPARAM lParam) {

	char promptMessage[PROMPT_MESSAGE_LEN];

	switch(iMsg)
	{
	case WM_INITDIALOG:
		sprintf(promptMessage,
			"Failed to open drive. Insert a disk into drive %c: and pr"
			"ess OK to continue.\n\n"
			"Caution: Any existing data on the disk "
			"you insert will be erased when the disk image is written "
			"to it.", targetDrive);		
		SendDlgItemMessage(hwnd,
			IDC_CONF_LABEL,
			WM_SETTEXT,
			0,
			(LPARAM)(LPCTSTR)&promptMessage);
		return TRUE;


	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
			case IDOK:
				EndDialog(hwnd, TRUE);
				return TRUE;
			case IDCANCEL:
				EndDialog(hwnd, FALSE);
				return TRUE;
		}
		break;
	}

	return FALSE;
}

LRESULT CALLBACK ConfirmDlgProc(HWND hwnd,
							 UINT iMsg,
							 WPARAM wParam,
							 LPARAM lParam) {

	char promptMessage[PROMPT_MESSAGE_LEN];

	switch(iMsg)
	{
	case WM_INITDIALOG:
		sprintf(promptMessage,
			"Insert a blank preformatted disk into drive %c: and press"
			" OK to continue. The disk must be of the same type as the"
			" disk image.\n\nCaution: Any existing data on the disk "
			"you insert will be erased when the disk image is written "
			"to it.", targetDrive);
		SendDlgItemMessage(hwnd,
			IDC_CONF_LABEL,
			WM_SETTEXT,
			0,
			(LPARAM)(LPCTSTR)&promptMessage);
		return TRUE;


	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
			case IDOK:
				EndDialog(hwnd, TRUE);
				return TRUE;
			case IDCANCEL:
				EndDialog(hwnd, FALSE);
				return TRUE;
		}
		break;
	}

	return FALSE;
}

void showErrorMessage(int code) {
	switch (code) {
	case VAMERR_SUCCESS:
		MessageBox(NULL,TEXT("No Error."),
						TEXT("!Error"), MB_OK | MB_ICONINFORMATION);
		return;
	case VAMERR_DRIVE_OPEN:
		MessageBox(NULL,TEXT("Drive is already open. This is likely a bug."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return;
	case VAMERR_OPEN_FAILED:
		MessageBox(NULL,TEXT("Failed to open drive."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_LOCK_FAILED:
		MessageBox(NULL,TEXT("Failed to lock drive for exclusive access."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_UNLOK_FAILED:
		MessageBox(NULL,TEXT("Failed to unlock drive for unexclusive access."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_DISMT_FAILED:
		MessageBox(NULL,TEXT("Failed to dismount volume."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_CLOSE_FAILED:
		MessageBox(NULL,TEXT("Failed to close drive."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_BAD_POINTER:
		MessageBox(NULL,TEXT("Bad Pointer. This is a bug."),
				TEXT("Error"), MB_OK | MB_ICONSTOP);
		return;
	case VAMERR_BAD_FORMAT:
		MessageBox(NULL,TEXT("Bad disk format."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_MEM_ALLOC_FAILED:
		MessageBox(NULL,TEXT("Unable to allocate memory. Possibly out of memory."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_DEV_WRITE_FAILED:
		MessageBox(NULL,TEXT("Writing to drive failed."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_DEV_READ_FAILED:
		MessageBox(NULL,TEXT("Reading from drive failed."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_DEV_SEEK_FAILED:
		MessageBox(NULL,TEXT("Seek failed."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	case VAMERR_NOT_ENOUGH_DATA:
		MessageBox(NULL,TEXT("Less data was received than expected."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;


	case VAMERR_FAILURE:
	default:
		MessageBox(NULL,TEXT("Unspecified Error."),
				TEXT("Error"), MB_OK | MB_ICONEXCLAMATION);
		return;
	}
}
