/*****************************************************************************
 * Application Name: WinVAM32 - Volume Access Module for 32bit Windows
 *        Copyright: (C) Copyright David Goodwin, 2010
 *          License: GNU GPL v2
 *****************************************************************************
	This file is part of "WinDIMG32".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 */
#define BUILD_DLL


/* For exported functions: */
#include "winvam32_dll.h"

/* For local functions */
#define V32_NOEXTGBL
#include "winvam32.h"
#undef V32_NOEXTGBL
#include "geom.h"
#include "vam_4x.h"
#include "vam_nt.h"

/* Windows headers */
#include <windows.h>
#include <winioctl.h>
#include <stdio.h>

/*
 * Global variables
 */
BOOL volume_is_open;
HANDLE hDrive;
vam_diskgeom diskgeom;
BOOL geom_valid;
BOOL isWindowsNT;

BOOL APIENTRY DllMain(HANDLE hModule, /* Handle to DLL Module */
					  DWORD ul_reason_for_call, /* Reason for Call */
					  LPVOID lpReserved ) /* Reserved */
{
	
	/* If we are attaching, make sure we are doing so on a supported operating
	   system. */
	if (ul_reason_for_call == DLL_PROCESS_ATTACH ||
		ul_reason_for_call == DLL_THREAD_ATTACH) {
		
		OSVERSIONINFO osvi;
		ZeroMemory(&osvi, sizeof(osvi));
		osvi.dwOSVersionInfoSize = sizeof(osvi);

		GetVersionEx((OSVERSIONINFO*)&osvi);

		if (osvi.dwPlatformId != VER_PLATFORM_WIN32_NT) {
			isWindowsNT = FALSE;

			// We are not on Windows NT: We only support NT right now.
			MessageBox(0,TEXT("Your Operating System is not supported by ")
				TEXT("this Volume Access Module. Unable to initialise."),
				TEXT("Unsupported Operating System"),
				MB_ICONERROR);

			return FALSE; 
		} else {
			isWindowsNT = TRUE;
		}
	}
	
	switch ( ul_reason_for_call )
	{
		case DLL_PROCESS_ATTACH:
			/* A process is loading the DLL. */
			volume_is_open = FALSE;
			hDrive = INVALID_HANDLE_VALUE;
			geom_valid = FALSE;
		break;
		case DLL_THREAD_ATTACH:
			/* A process is creating a new thread. */
		break;
		case DLL_THREAD_DETACH:
			/* Thread has detached */
		break;
		case DLL_PROCESS_DETACH:
			/* The DLL has been unloaded */
			vam_close_volume();
			CloseHandle(hDrive);
		break;
	}
	return TRUE;
}


EXPORT int vam_open_volume(char drive_letter) 
{
	/*
	 * Example volume name: \\.\A:
	 */
	char volume_name[10];
	
	sprintf(volume_name,"\\\\.\\%c:",drive_letter);
	
	/* If a drive is already open, fail. */
	if (volume_is_open) {
		return VAMERR_DRIVE_OPEN;
	}


	hDrive = INVALID_HANDLE_VALUE;
	hDrive = CreateFileA(volume_name,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_SEQUENTIAL_SCAN |
		FILE_FLAG_NO_BUFFERING |
		FILE_ATTRIBUTE_SYSTEM,
		NULL);

	if (hDrive == INVALID_HANDLE_VALUE) {
		return VAMERR_OPEN_FAILED;
	}

	if (!lock_volume(hDrive)) {
		return VAMERR_LOCK_FAILED;
	}

	volume_is_open = TRUE;

	// We really should handle this differently.
	return VAMERR_SUCCESS;
}

EXPORT int vam_close_volume() {
	/* Try to dismount the volume */
	BOOL ret = dismount_volume(hDrive);
	if (!ret) return VAMERR_DISMT_FAILED;

	/* Try to unlock the volume */
	ret = unlock_volume(hDrive);
	if (!ret) return VAMERR_UNLOK_FAILED;

	/* Close the handle */
	ret = CloseHandle(hDrive);
	if (!ret) return VAMERR_CLOSE_FAILED;

	volume_is_open = FALSE;
	hDrive = INVALID_HANDLE_VALUE;
	geom_valid = FALSE;

	return VAMERR_SUCCESS;
}

EXPORT BOOL vam_is_drive_open() {
	return volume_is_open;
}

EXPORT vam_diskgeom vam_get_geometry() {
    return diskgeom;
}

EXPORT void vam_set_geometry(vam_diskgeom dg) {
	diskgeom = dg;
}

EXPORT int vam_set_geometry_from_data(const char * bootsect, int bootsect_length) {
	int e;

	e = get_dosgeom(&diskgeom,bootsect,bootsect_length);

	return e;
}

EXPORT int vam_write_data(long sector, 
						  const char * data, 
						  int data_length, 
						  DWORD *dwBytesWritten) {
	/*
	 * The method for drive access differs under Windows 4.x (95/98/ME) and
	 * Windows NT. So we must call the appropriate function.
	 */
	if (isWindowsNT)
		return write_data_nt(sector,data,data_length,dwBytesWritten);
	else
		return write_data_4(sector,data,data_length,dwBytesWritten);
}



EXPORT int vam_read_data(long track,
						 char* data,
						 int data_length,
						 int *length_read) {
	if (isWindowsNT) 
		return read_data_nt(track,data,data_length,length_read);
	else
		return read_data_4(track,data,data_length,length_read);

}

EXPORT int vam_get_version() {
	/* Version 1.0.0 */
	return 100;
}

BOOL dismount_volume(HANDLE hDisk) {

	DWORD returned_byte_count;
	int ret = DeviceIoControl(hDisk,
		FSCTL_DISMOUNT_VOLUME,
		NULL,
		0,
		NULL,
		0,
		&returned_byte_count,
		NULL);

	if (ret == 0) return FALSE;
	else return TRUE;
}

BOOL lock_volume(HANDLE hDisk) {


	DWORD returned_byte_count;
	int ret = DeviceIoControl(hDisk,
		FSCTL_LOCK_VOLUME,
		NULL,
		0,
		NULL,
		0,
		&returned_byte_count,
		NULL);

	if (ret == 0) return FALSE;
	else return TRUE;
}

BOOL unlock_volume(HANDLE hDisk) {

	DWORD returned_byte_count;
	int ret = DeviceIoControl(hDisk,
		FSCTL_UNLOCK_VOLUME,
		NULL,
		0,
		NULL,
		0,
		&returned_byte_count,
		NULL);

	if (ret == 0) return FALSE;
	else return TRUE;
}

EXPORT int vam_read_diskgeom() {
	if (isWindowsNT) 
		return read_diskgeom_nt();
	else
		return read_diskgeom_nt();
}
